%-----------------------------------------------------------------------------
% Original Quintus Prolog WordNet 1.6 Version:
%   (C) Sandiway Fong, November 23rd 2000. NEC Research Institute, Inc.
% Sicstus Prolog WordNet 1.7.1 version
%   (C) Sandiway Fong, October 27th 2003. University of Arizona
%   Notes:
%	Reindexes synset_ids only
%	Not compatible with QP WN 1.6, which indexed by synset_id,offset
%-----------------------------------------------------------------------------
% To load this program, do:
%	?- restore(connection).
%
% connect(W1,W2)
%
%   Search program to display the connection between two words (W1 and W2) 
%   in Wordnet.
%   A word can be a Prolog atom or Word/C, C a category label
%	e.g. ?- connect(happy,unhappy).
%	     ?- connect(happy/a,unhappy/a).
%   WordNet defines categories {n,v,a,s,r}
%	s  satellite adjective
%	r  adverb
%   For (a) shortest path solution only, set counter 15 to be non-zero
%	e.g. ?- ctr_set(15,1).
%   For synset ids to be included in the report, set counter 16  be non-zero
%   Each relation can be individually switched on/off by commenting out
%	its link/3 clause. e.g. ihyp/2 is called here
%	% inverse instance of
%	link(I1,I2,2621440) :-		% 2^19+2^21
%		ihyp(I1,I2).
%   File g.pl (defines gloss) is also included but not loaded by default
%   Note: g.ql is a 21Mb file
%   To load it, do:
%	?- ensure_loaded(g).
%
% Stage 7 implementation: 
% 1. uses global array for loop detection
% 2. encode link as single number (Id + link type)
%	111223 synsets in WN 1.7.1, Use lower 17 bits to encode Id
%	(2^17-1 = 131072)
%	bits 18 (2^17) and above used by link type
%	bit 32 (2^31) is reserved for marking
% 3. uses an optimized array for the path tree
% 4. Id encodes synset_id only, offset where needed in special link bits

% Wordnet databases

:- ensure_loaded(ant).		% ant(I1,O1,I2,O2)
:- ensure_loaded(at).		% at(I1,I2) (adjs and nouns)
:- ensure_loaded(cs).		% cs(I1,I2) (verbs)
:- ensure_loaded(ent).		% ent(I1,I2) (verbs)
% :- ensure_loaded(g).		% g(S,Gloss)
:- ensure_loaded(hyp).		% hyp(I1,I2) (nouns and verbs)
:- ensure_loaded(ihyp).		% ihyp(I1,I2) (nouns and verbs)
:- ensure_loaded(mm).		% mm(I1,I2) (nouns)
:- ensure_loaded(mp).		% mp(I1,I2) (nouns)
:- ensure_loaded(ms).		% ms(I1,I2) (nouns)
:- ensure_loaded(per).		% per(I1,O1,I2,O2)
:- ensure_loaded(ppl).		% ppl(I1,O1,I2,O2)
:- ensure_loaded(sa).		% sa(I1,O1,I2,O2)
:- ensure_loaded(sim).		% sim(I1,I2) (adjs)
:- ensure_loaded(vgp).		% vgp(I1,I2) (verbs)
:- ensure_loaded(s2).		% s2(Word,C,I,S,O)
:- ensure_loaded(s3).		% s3(I,O,S,Word,C)
:- ensure_loaded(s4).		% s4(I,S,List)
:- ensure_loaded(file).		% file(S,FN) FN file number
:- ensure_loaded(fn).		% fn(I1,O1,I2,O2)
:- ensure_loaded(lexnames).	% ln(FN,LN,C) (Lex name)

% does a breadth-first search

connect(W1,W2) :-
	initEndMsg,
	in_db(W1),
	in_db(W2),
	on_exception(done,connect1(W1,W2),true),
	fail.
connect(_,_) :-
	displayEndMsg.

connect1(W1,W2) :- 
	reset_mark,
	reset_pt,
	lookUpAllAndMark(W1),
	lookUpAll(W2,Id2s),
	add_sep,
	zip_pt,
	connectL(Id2s,2),
	step(Id2s,3).

step(Id2s,J) :-
	ctr_set(13,0),
	add_sep,
	zip_pt,
	connectL(Id2s,J),
	K is J + 1,
	!,					  % green
	(ctr_is(13,0)
	-> true
	;   step(Id2s,K)).

connectL(Id2s,J) :- 
	at_pt(V),
	(V == 0
	-> true
	;  Id is V /\ 131071,	% bottom 17 bits
	   connect(Id,Id2s,J),
	   !,					  % green
	   connectL(Id2s,J)).

% find connections from Id to new nodes encoded in Links
% Id2s: list of terminating nodes
% J: level of breath-first search

connect(Id,Id2s,J) :-
	at2_pt(N),
	linkLD(Id,N,J,Id2s),
	fail.
connect(_,_,_).

% link, check loop and check done
linkLD(Id,N,J,Id2s) :-
	link(Id,Id2,Code),
	noLoop(Id2,J),
	Link is Id2 + Code,
	notDone(Id2,Link,N,Id2s).

% ctr 16 controls display format, and ctr 15 whether to stop at one answer
notDone(Id,Link,N,Ids) :-
	memberFP(Id,O,Ids)
	-> (ctr_is(16,0)
	   -> displayPath(Link,N,Id,O)
	   ;  displayPathD(Link,N,Id,O)),
	   displayNodes,
	   (ctr_is(15,0)
	   -> fail
	   ;   raise_exception(done))
	;  add_pt(Link,N),
	   ctr_set(13,1).

% e.g. node marked at level 3, only level 3 (or lower) can use the node
noLoop(Id,U) :-
	get_mark(Id,V),
	(V == 0
	-> set_mark(Id,U)
	;  U =< V).

% simple operator links
% Encoding:
% 2^0 - 2^16 (17 bits)  Id (enough for 131071 synsets)
% 2^17 (bit 18) 0 signal simple operator decoding
% 2^18 upwards - operator encoded as follows
%	sim 1, hyp 10, at 11, cs 100, ent 101, mm 110, ms 111, mp 1000
% 	vgp 1001, ihyp 1010
% Note: bit 32 (2^31) reserved for MARK, Max +int is 2^31-1

% similar
link(I1,I2,262144) :-		% 2^18
	sim(I1,I2).

% instance of
link(I1,I2,524288) :-		% 2^19
	hyp(I1,I2).

% (adj) value of (noun)
link(I1,I2,786432) :-		% 2^18+2^19
	at(I1,I2).

% (verb) caus (verb)
link(I1,I2,1048576) :-		% 2^20
	cs(I1,I2).

% (verb) entails (verb)
link(I1,I2,1310720) :-		% 2^18+2^20
	ent(I1,I2).

% (noun) member meronym of (noun)
link(I1,I2,1572864) :-		% 2^19+2^20
	mm(I1,I2).

% (noun) substance meronym of (noun)
link(I1,I2,1835008) :-		% 2^18+2^19+2^20
	ms(I1,I2).

% (noun) part meronym of (noun)
link(I1,I2,2097152) :-		% 2^21
	mp(I1,I2).

% (verbs) are grouped together
link(I1,I2,2359296) :-		% 2^18+2^21
	vgp(I1,I2).

% inverse instance of
link(I1,I2,2621440) :-		% 2^19+2^21
	ihyp(I1,I2).

% operator/operand links
% Encoding:
% 2^0 - 2^16 (17 bits)  Id (enough for 131071 synsets)
% 2^17 1 signal operator/operand decoding
% 2^18 - 2^20 8 codes
% 	ant 000, fn 001, ppl 010, per 011, sa 100
%	101, 110 free
%	start 111 (takes single operand) - aka 1966080
% 2^21 - 2^30 (10 bits, 2 operands 0-27)
% 2^21 - 2^25 (5 bits) - 31
%	ant 10, fn 23, ppl 3, per 8, sa 7
% 2^26 - 2^30 (5 bits) - 31
%	ant 10, fn 23, ppl 9, per 22, sa 11
% 2^31 (1 bit)  (reserved) mark


% antonyms
link(I1,I2,V) :-				  % 2^17+O1+O2
	ant(I1,O1,I2,O2),
	V is 131072 + O1 << 21 + O2 << 26.

% common file number
link(I1,I2,V) :-				  % 2^17+2^18+O1+O2
	fn(I1,O1,I2,O2),
	V is 393216 + O1 << 21 + O2 << 26.

% (adj) participle of (verb)
link(I1,I2,V) :-				  % 2^17+2^19+O1+O2
	ppl(I1,O1,I2,O2),
	V is 655360 + O1 << 21 + O2 << 26.

% 1: adj 2: noun/adj (adj pertains to noun/adj)
% 1: adv 2: adj      (adv pertains to adj)
link(I1,I2,V) :-				  % 2^17+2^18+2^19+O1+O2
	per(I1,O1,I2,O2),
	V is 917504 + O1 << 21 + O2 << 26.

% see also
link(I1,I2,V) :-				  % 2^17+2^20+O1+O2
	sa(I1,O1,I2,O2),
	V is 1179648 + O1 << 21 + O2 << 26.

lookUpAll(W,Ids) :- findall([Id,O],lookUp(W,Id,O),Ids).

lookUpAllAndMark(W) :- 
	lookUpAndMark(W),
	fail.
lookUpAllAndMark(_).

lookUp(W,I,O) :-
	W = W1/C
	-> s2(W1,C,I,_,O)
	;  s2(W,_,I,_,O).

lookUpAndMark(W) :-
	W = W1/C
	-> s2(W1,C,I,_,O),
	   set_mark(I,1),
	   mkStartLink(I,O,V),
	   add_pt(V,0)
	;  s2(W,_,I,_,O),
	   set_mark(I,1),
	   mkStartLink(I,O,V),
	   add_pt(V,0).

mkStartLink(I,O,V) :-		% start: 2^17+2^18+2^19+2^20+O+I
	   V is I + 1966080 + O << 21.

in_db(W) :- 
	W = W1/C
	-> in_db1(W1,C)
	;  in_db1(W).

in_db1(W)   :- s2(W,_,_,_,_), !.
in_db1(W)   :- fail([W,' not found!']).
in_db1(W,C) :- s2(W,C,_,_,_), !.
in_db1(W,C)   :- fail([W/C,' not found!']).

%%% Display Code
%
% displayPath(Path)
%
%   displays the Path.
%   Path is a list of Links.
%   Note: assume Path is in reverse order
%   Note: keeps a count of the number of Paths, initialization: initEngMsg
%
% displayPathD(Path) 
%
%   detailed version of above, includes synset numbers and offsets
%
% displayLink(Link)
%
%   displays the Link.
%   Link possibilities are:	Relation:
%     ppl(Id)			particle of Id
%     ant(Id)			antonym of Id
%     sim(Id)			similar to Id
%     per(Id)			pertains to Id
%     sa(Id)			see also Id
%     hyp(Id)			hypernym of Id
%     ihyp(Id)			hyponym of Id
%     syn(Id)			same synset as Id
%     fn(Id)			same word as (related by file number)
%
% displayLinkD(Link) 
%
%   detailed version of above, includes synset numbers and offsets
%

displayPath(Link,N,Id,O) :- 
	extractP(N,Path),
	displayPath1([Link|Path],_,_),
	displayEnd(Link,Id,O).

displayPath1([],_,_) :- ctr_inc(0).
displayPath1([H|T],I,U) :-
	displayPath1(T,J,V),
	!,			% red
	displayLink(H,J,V,I,U).	

% for detailed display

displayPathD(Link,N,Id,O) :- 
	extractP(N,Path),
	displayPathD1([Link|Path],_,_),
	displayEndD(Link,Id,O).

extractP(N,[V|P]) :-
	bk_pt(N,V),
	bk2_pt(N,M),
	(M == 0
	-> P = []
	;  extractP(M,P)).

displayPathD1([],_,_) :- ctr_inc(0).
displayPathD1([H|T],I,U) :-
	displayPathD1(T,J,V),
	!,			% red
	displayLinkD(H,J,V,I,U).

% we end the search by matching the same synset
% however, ending item may not be the same as the target
% hence, an extra step "in same synset" may need to be displayed

displayEnd(Link,I,O) :-
	sameItem(Link,I,O),
	!.			% red
displayEnd(_,I,O) :-
	s3(I,O,_,W,C),
	s4(I,_,L),
	format('~w/~w is in the synset ~w~N',[W,C,L]).

displayEndD(Link,I,O) :-
	sameItem(Link,I,O),
	!.			% red
displayEndD(_,I,O) :-
	s3(I,O,SS,W,C),
	s4(I,_,L),
	format('~w/~w (~d,~d) is in the synset ~w~N',[W,C,SS,O,L]).

% Matching cases
% Op(_,_,O)  direct match
% Op(I) synset I has only one element
sameItem(Link,_,O) :-
	decode(Link,Op),
	functor(Op,_,N),
	sameItem1(N,Op,O).

sameItem1(3,Op,O) :-
	arg(3,Op,O).
sameItem1(1,Op,_) :-
	arg(1,Op,I),
	once(s4(I,_,[_])).	% green
	    
%%% displayLink(K,_,_)
%%% displayLinkD(K,_,_)	reports WordNet synset_id,offset pairs as well

% for simple report
displayLink(K,I,U,J,V) :-
	decode(K,Link),
	displayLink1(Link,I,U,J,V).

% for a more detailed report - including WN synset_id,offset pairs
displayLinkD(K,I,U,J,V) :-
	decode(K,Link),
	displayLinkD1(Link,I,U,J,V).

% unpacking integer representation
decode(K,Link) :-		% switch
	X is K /\ 131072,
	decode1(X,K,Link).

decode1(0,K,Link) :-		% no operands
	I is K /\ 131071,	% 2^17 - 1
	U is K >> 17,
	linkRel(U,R),
	Link =.. [R,I].
decode1(131072,K,Link) :-	% operator/operand decode
	U is K /\ 1966080,	% mask bits 17-20
	I is K /\ 131071,	% 2^17 - 1
	O1 is (K /\ 65011712) >> 21, % mask bits 21-25
	O2 is (K /\ 2080374784) >> 26, % mask bits 26-30
	linkRelOO(U,R),
	Link =.. [R,I,O1,O2].

% operator/operand link relations
linkRelOO(131072,ant).		% bit 17, 000
linkRelOO(393216,fn). 		% bit 17, 001
linkRelOO(655360,ppl).		% bit 17, 010
linkRelOO(917504,per).		% bit 17, 011
linkRelOO(1179648,sa).		% bit 17, 100
linkRelOO(1966080,start).	% bit 17, 111

% simple operator relations
%	sim 1, hyp 10, at 11, cs 100, ent 101, mm 110, ms 111, mp 1000
%	vgp 1001, ihyp 1010
linkRel(2,sim).			% 2^18
linkRel(4,hyp).			% 2^19
linkRel(6,at).			% 2^18+2^19
linkRel(8,cs).			% 2^20
linkRel(10,ent).		% 2^20+2^18
linkRel(12,mm).			% 2^20+2^19
linkRel(14,ms).			% 2^20+2^19+2^18
linkRel(16,mp).			% 2^21
linkRel(18,vgp).		% 2^18+2^21
linkRel(20,ihyp).		% 2^19+2^21

% synset_id,offset links
displayLinkD1(start(I,O,_),_,_,I,ss) :-
	s3(I,O,SS,W,C),
	s4(I,_,L),
	format('~n~w/~w (~d,~d) is in synset ~w~N',[W,C,SS,O,L]).

displayLinkD1(ppl(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,SS,W,C),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format(' is a particle of ~w/~w (~d,~d)~N',[W,C,SS,O2]).

displayLinkD1(ant(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,SS,W,C),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format(' and ~w/~w (~d,~d) are antonyms~N',[W,C,SS,O2]).

displayLinkD1(per(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,SS,W,C),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format('~w pertains to ~w/~w (~d,~d)~N',[PW,W,C,SS,O2]).

displayLinkD1(sa(I,0,0),J,_,I,ss) :-
	!,			% red
	s4(I,SS,L),
	s4(J,_,PL),
	format('For ~w, see also synset ~w (~d)~N',[PL,L,SS]).

displayLinkD1(sa(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,SS,W,C),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format(', see also ~w/~w (~d,~d)~N',[W,C,SS,O2]).

displayLinkD1(fn(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,SS,W,C),
	file(SS,FN),
	ln(FN,Type,_),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format(' and ~w/~w (~d,~d) related by ~w~N',[W,C,SS,O2,Type]).
	    
% synset links
displayLinkD1(sim(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w and ~w (~d) are similar~N',[PL,List,SS]).

displayLinkD1(hyp(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w is an instance of ~w (~d)~N',[PL,List,SS]).

displayLinkD1(ihyp(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w (~d) is an instance of ~w~N',[List,SS,PL]).

displayLinkD1(at(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w is an attribute of ~w (~d)~N',[PL,List,SS]).

displayLinkD1(cs(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w causes ~w (~d)~N',[PL,List,SS]).

displayLinkD1(ent(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w entails ~w (~d)~N',[PL,List,SS]).

displayLinkD1(mm(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w is a member holonym of ~w (~d)~N',[PL,List,SS]).

displayLinkD1(mp(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w is a part holonym of ~w (~d)~N',[PL,List,SS]).

displayLinkD1(ms(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w substance holonym of ~w (~d)~N',[PL,List,SS]).

displayLinkD1(vgp(I),J,_,I,ss) :- 
	s4(I,SS,List),
	s4(J,_,PL),
	format('~w is grouped with ~w (~d)~N',[PL,List,SS]).

% doesn't report synset_id,offset pairs
displayLink1(start(I,O,_),_,_,I,ss) :-
	s3(I,O,_,W,C),
	s4(I,_,L),
	format('~n~w/~w is in ~w~N',[W,C,L]).

displayLink1(ppl(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,_,W,C),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format(' is a particle of ~w/~w~N',[W,C]).

displayLink1(ant(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,_,W,C),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format(' and  ~w/~w are antonyms~N',[W,C]).

displayLink1(per(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,_,W,C),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format(' pertains to ~w/~w~N',[W,C]).

displayLink1(sa(I,0,0),J,_,I,ss) :-
	!,			% red
	s4(I,_,L),
	s4(J,_,PL),
	format('For ~w, see also synset ~w~N',[PL,L]).

displayLink1(sa(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,_,W,C),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format(', see also ~w/~w~N',[W,C]).

displayLink1(fn(I,O1,O2),J,V,I,s) :- 
	s3(I,O2,SS,W,C),
	file(SS,FN),
	ln(FN,Type,_),
	s3(J,O1,_,PW,_),
	condFormat(V,PW,J),
	format(' and ~w/~w related by ~w~N',[W,C,Type]).

% synset links
displayLink1(sim(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w and ~w are similar~N',[PL,List]).

displayLink1(hyp(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w is an instance of ~w~N',[PL,List]).

displayLink1(ihyp(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w is an instance of ~w~N',[List,PL]).

displayLink1(at(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w is an attribute of ~w~N',[PL,List]).

displayLink1(cs(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w causes ~w~N',[PL,List]).

displayLink1(ent(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w entails ~w~N',[PL,List]).

displayLink1(mm(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w is a member holonym of ~w~N',[PL,List]).

displayLink1(mp(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w is a part holonym of ~w~N',[PL,List]).

displayLink1(ms(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w is a substance holonym of ~w~N',[PL,List]).

displayLink1(vgp(I),J,_,I,ss) :- 
	s4(I,_,List),
	s4(J,_,PL),
	format('~w is grouped with ~w~N',[PL,List]).

% s means single element write last time around
% condFormat/3 is called from a single element write this time around as well
% if single element write twice in a row, write synset out too
condFormat(V,W,I) :-
	V == s ->
	s4(I,_,L),
	format('~w in ~w',[W,L])
	; write(W).

% requires ctr

initEndMsg :- ctr_set(0,0).

displayEndMsg :-
	ctr_is(0,V),
	displayEndMsg(V),
	displayNodes.

displayNodes :-
	count_marks(N),
	format('~N#nodes visited: ~d~N',[N]).

displayEndMsg(N) :- displayEndMsg1(N), !.	  % red

displayEndMsg1(0) :- format('~NNo path',[]).
displayEndMsg1(1) :- format('~NUnique path',[]).
displayEndMsg1(N) :- format('~NReachable in ~d different ways',[N]).

fail(L) :-
	writeL(L),
	fail.

writeL([]) :- nl.
writeL([H|T]) :- write(H), !, writeL(T).	  % red

writeL(S,[]) :- nl(S).
writeL(S,[H|T]) :- write(S,H), !, writeL(S,T).	  % red

% Following Quintus member/2 with loop unrolling for 10%~20% gain
memberFP(X,O, [[X,O]|_]    ).
memberFP(X,O, [_,[X,O]|_]  ).
memberFP(X,O, [_,_,[X,O]|_]).
memberFP(X,O, [_,_,_|L]) :- memberFP(X,O, L).

%% C interface
% Sicstus Prolog version

foreign_resource(connection, [reset_mark,set_mark,get_mark,count_marks,
			      reset_pt,add_sep,add_pt,zip_pt,at_pt,at2_pt,
			      bk_pt,bk2_pt,print_pt]).


% Quintus Prolog version

% foreign_file('connection.o', [reset_mark,set_mark,get_mark,count_marks,
%			      reset_pt,add_sep,add_pt,zip_pt,at_pt,at2_pt,
%			      bk_pt,bk2_pt,print_pt]).

foreign(reset_mark,  reset_mark).
foreign(set_mark,    set_mark(+integer,+integer)).
foreign(get_mark,    get_mark(+integer,[-integer])).
foreign(count_marks, count_marks([-integer])).
foreign(reset_pt,    reset_pt).
foreign(add_sep,     add_sep).
foreign(zip_pt,      zip_pt).
foreign(add_pt,      add_pt(+integer,+integer)).
foreign(at_pt,       at_pt([-integer])).
foreign(at2_pt,      at2_pt([-integer])).
foreign(bk_pt,       bk_pt(+integer,[-integer])).
foreign(bk2_pt,      bk2_pt(+integer,[-integer])).
foreign(print_pt,    print_pt).

% Sicstus Prolog version

:- load_foreign_resource(connection).

% Quintus Prolog version

%:- load_foreign_files('connection.o',[]),
%	fail
%	;  abolish([foreign_file/2,foreign/2]).

:- ensure_loaded(ctr).


